﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace ParameterSubstitution
{
    public static class Substitution
    {
        public static string FormatByName(this string s, Dictionary<string,object> parameterdictionary)
        {
            return KeyValueReplacer(s, parameterdictionary);
        }

        public static string FormatByName(this string s, object parameterclass)
        {
            var d = PropertiesToDictionary(parameterclass);
            return KeyValueReplacer(s, d);
        }

        private static Dictionary<string, object> PropertiesToDictionary(object o)
        {
            var t = o.GetType();

            if (!t.IsClass)
                throw new ArgumentException("Can only resolve properties in classes.");

            var p = t.GetProperties();
            var propcollection = p.ToDictionary(propertyInfo => propertyInfo.Name, propertyInfo => propertyInfo.GetValue(o, null));
            return propcollection;
        }

        private static string KeyValueReplacer(string template, Dictionary<string, object> parametercollection)
        {
            // Get the parameters from the template

            Regex m_regexmatch = new Regex(@"(?<start>\{)+(?<property>[\w\.\[\]]+)(?<format>:[^}]+)?(?<end>\})+",
                RegexOptions.Compiled | RegexOptions.CultureInvariant | RegexOptions.IgnoreCase);

            var matches = m_regexmatch.Matches(template).OfType<Match>();

            // Get the parameters { to } and index of any suffixed : formatting code
            var paramlist1 = matches.Select(s => new { Parameter = s.Value, FormatIndex = s.Value.IndexOf(':') });

            // Extract the key (parameter name) and formatting code from template
            var paramlist2 =
                paramlist1.Select(s =>
                                  new
                                  {
                                      Key = s.Parameter.Substring(1, s.FormatIndex < 1 ? s.Parameter.Length - 2 : s.FormatIndex - 1),
                                      Format = s.FormatIndex < 1 ? "" : s.Parameter.Substring(s.FormatIndex, s.Parameter.Length - s.FormatIndex - 1),
                                      s.Parameter,
                                  });

            // Perform existence checks from both sides
            var paramlist2unique = paramlist2.Select(s => s.Key).Distinct();
            foreach (var key in parametercollection.Keys)
                if (!paramlist2unique.Contains(key))
                    throw new ArgumentException("Parameter [" + key + "] not found in template");

            foreach (var key in paramlist2unique)
                if (!parametercollection.ContainsKey(key))
                    throw new ArgumentException("Template variable {" + key + "} not found in parameter collection");

            // Join the template parameters and the values from parametercollection
            var paramlist3 =
                from p in paramlist2
                from v in parametercollection
                where p.Key == v.Key
                select new
                {
                    //p.Key,
                    //p.Format,
                    p.Parameter,
                    //v.Value,
                    ValueString = String.Format("{0" + p.Format + "}", v.Value)
                };

            // Replace the values into the template
            var sb = new StringBuilder(template);
            foreach (var p in paramlist3)
                sb.Replace(p.Parameter, p.ValueString);

            return sb.ToString();
        }
    }
}
